from threading import Condition, RLock, Thread
from typing import Any, Callable

import numpy as np
from numpy.typing import NDArray


class Batcher:
    """Background thread-safe worker to collect incoming elements to
    batches and performing callback on them.
    """

    def __init__(
        self,
        size: int,
        timeout: float,
        callback: Callable[[NDArray[Any]], Any]
    ) -> None:
        self._size = size
        self._timeout = timeout
        self._callback = callback

        self._batch: list[Any] = []

        self._is_waiting_first_element = False
        self._is_flushed = False
        self._is_finished = False

        self._lock = RLock()
        self._first_element_condition = Condition(self._lock)
        self._size_condition = Condition(self._lock)

        self._thread = Thread(target=self._run_cycle)
        self._thread.start()

    def _run_cycle(self):
        """Target method for thread that tracks conditions."""
        while True:
            with self._lock:
                self._is_waiting_first_element = True
                self._is_flushed = False
                self._is_finished = False

                self._first_element_condition.wait()

                if self._is_finished:
                    return

                self._size_condition.wait(self._timeout)

                if self._is_finished:
                    return

                if self._is_flushed:
                    continue

                batch = self._batch
                self._batch = []

            self._flush_batch(batch)

    def _flush_batch(self, batch):
        """Perform callback on current batch."""
        if batch:
            self._callback(np.array(batch))

    def add(self, element: Any) -> None:
        """Add element to current batch."""

        complete_batch = None

        with self._lock:
            self._batch.append(element)

            if len(self._batch) >= self._size:
                complete_batch = self._batch
                self._batch = []

                self._is_flushed = True
                self._size_condition.notify_all()
            elif self._is_waiting_first_element:
                self._is_waiting_first_element = False
                self._first_element_condition.notify_all()

        if complete_batch:
            self._flush_batch(complete_batch)

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_value, traceback):
        with self._lock:
            self._is_finished = True
            self._first_element_condition.notify_all()
            self._size_condition.notify_all()

        self._thread.join()
        self._flush_batch(self._batch)



import importlib
from abc import ABC
from typing import Any, Literal

from eventum_plugins.input.base import BaseInputPlugin
from eventum_plugins.output.base import BaseOutputPlugin
from eventum_plugins.utils.modules import get_module_names
from pydantic import BaseModel, Field, create_model, model_validator


class MutexFieldsModel(ABC, BaseModel, extra='forbid', frozen=True):
    @model_validator(mode='after')
    def validate_mutual_exclusion(self):
        """Validate that only one field is passed and rest fields is `None`."""
        values = [
            self.__getattribute__(field) for field in self.model_fields.keys()
        ]

        if values.count(None) == (len(values) - 1):
            return self

        raise ValueError('Only one key can be defined at this level')

    def get_name(self) -> str:
        """Get name of used parameter."""
        return list(filter(
            lambda attr: self.__getattribute__(attr) is not None,
            self.model_fields.keys()
        )).pop()

    def get_value(self) -> Any:
        """Get value if used parameter."""
        return self.__getattribute__(self.get_name())


def get_plugins_to_config_mapping(
    plugin_type: Literal['input', 'output']
) -> dict[str, BaseModel]:
    """Get mapping where keys are a plugin names and values are config
    classes for specified types of plugins."""
    mapping: dict[str, BaseModel] = dict()
    for module_name in get_module_names(f'eventum_plugins.{plugin_type}'):
        module = importlib.import_module(
            name=f'eventum_plugins.{plugin_type}.{module_name}'
        )
        try:
            mapping[module_name] = module.CONFIG_CLASS
        except AttributeError:
            pass

    return mapping


def _load_plugin_class(
    plugin_type: Literal['input', 'output'],
    plugin_name: str
) -> type[BaseInputPlugin] | type[BaseOutputPlugin]:
    """Get plugin class. Raise `ValueError` if plugin with specified
    name is not found."""
    try:
        module = importlib.import_module(
            name=f'eventum_plugins.{plugin_type}.{plugin_name}'
        )
        return module.PLUGIN_CLASS
    except (ModuleNotFoundError, AttributeError):
        raise ValueError(
            f'{plugin_type.capitalize()} plugin "{plugin_name}" is not found'
        )


def load_input_plugin_class(plugin_name: str) -> type[BaseInputPlugin]:
    """Get input plugin class. Raise `ValueError` if plugin with
    specified name is not found."""
    return _load_plugin_class(      # type: ignore[return-value]
        plugin_type='input',
        plugin_name=plugin_name
    )


def load_output_plugin_class(plugin_name: str) -> type[BaseOutputPlugin]:
    """Get output plugin class. Raise `ValueError` if plugin with
    specified name is not found."""
    return _load_plugin_class(      # type: ignore[return-value]
        plugin_type='output',
        plugin_name=plugin_name
    )


InputConfigMapping: type[
    MutexFieldsModel
] = create_model(                   # type: ignore[call-overload]
    'InputConfigMapping',
    __base__=MutexFieldsModel,
    __cls_kwargs__={'frozen': True},
    **{
        plugin_name: (config_class, Field(None))
        for plugin_name, config_class in get_plugins_to_config_mapping(
            plugin_type='input'
        ).items()
    }
)

OutputConfigMapping: type[
    MutexFieldsModel
] = create_model(                   # type: ignore[call-overload]
    'OutputConfigMapping',
    __base__=MutexFieldsModel,
    __cls_kwargs__={'frozen': True},
    **{
        plugin_name: (config_class, Field(None))
        for plugin_name, config_class in get_plugins_to_config_mapping(
            plugin_type='output'
        ).items()
    }
)



import asyncio
import logging
import signal
from datetime import datetime
from multiprocessing import Queue
from multiprocessing.sharedctypes import SynchronizedBase
from multiprocessing.synchronize import Event as EventClass
from typing import Callable, NoReturn, Optional

import numpy as np
from eventum_plugins.event.base import (EventPluginConfigurationError,
                                        EventPluginRuntimeError)
from eventum_plugins.event.jinja import JinjaEventConfig, JinjaEventPlugin
from eventum_plugins.input.base import (InputPluginConfigurationError,
                                        InputPluginRuntimeError)
from eventum_plugins.output.base import (BaseOutputPlugin,
                                         OutputPluginConfigurationError,
                                         OutputPluginRuntimeError)
from jinja2 import BaseLoader
from numpy.typing import NDArray
from pytz import timezone
from setproctitle import getproctitle, setproctitle

from eventum_core.batcher import Batcher
from eventum_core.plugins_connector import (MutexFieldsModel,
                                            load_input_plugin_class,
                                            load_output_plugin_class)
from eventum_core.settings import Settings, TimeMode

logger = logging.getLogger(__name__)


def subprocess(name: str) -> Callable:
    """Parametrized decorator for all subprocesses."""

    def decorator(f: Callable):
        def wrapper(*args, **kwargs):
            setproctitle(f'{getproctitle()} [{name}]')

            signal.signal(signal.SIGINT, lambda signal, stack_frame: exit(0))

            result = f(*args, **kwargs)
            return result

        return wrapper
    return decorator


def _terminate_subprocess(
    is_done: EventClass,
    exit_code: int = 0,
    signal_queue: Optional[Queue] = None
) -> NoReturn:
    """Handle termination of subprocess."""
    if signal_queue is not None:
        signal_queue.put(None)
    is_done.set()
    exit(exit_code)


@subprocess('input')
def start_input_subprocess(
    config: MutexFieldsModel,
    settings: Settings,
    time_mode: TimeMode,
    queue: Queue,
    is_done: EventClass,
) -> None:
    plugin_name = config.get_name()
    input_conf = config.get_value()

    logger.info(f'Initializing "{plugin_name}" input plugin')

    try:
        plugin_class = load_input_plugin_class(plugin_name=plugin_name)
        input_plugin = plugin_class(
            config=input_conf,
            tz=timezone(settings.timezone)
        )
    except ValueError as e:
        logger.error(f'Failed to load input plugin: {e}')
        _terminate_subprocess(is_done, 1, queue)
    except InputPluginConfigurationError as e:
        logger.error(f'Failed to initialize input plugin: {e}')
        _terminate_subprocess(is_done, 1, queue)
    except Exception as e:
        logger.error(
            'Unexpected error occurred during initializing '
            f'input plugin: {e}'
        )
        _terminate_subprocess(is_done, 1, queue)

    logger.info('Input plugin is successfully initialized')

    try:
        with Batcher(
            size=settings.events_batch_size,
            timeout=settings.events_batch_timeout,
            callback=queue.put
        ) as batcher:
            plugin_mode = input_plugin.__getattribute__(time_mode.value)
            plugin_mode.__call__(on_event=batcher.add)
    except AttributeError:
        logger.error(
            f'"{plugin_name}" input plugin does not support "{time_mode}" mode'
        )
        _terminate_subprocess(is_done, 1, queue)
    except InputPluginRuntimeError as e:
        logger.error(f'Error occurred during input plugin execution: {e}')
        _terminate_subprocess(is_done, 1, queue)
    except Exception as e:
        logger.error(
            f'Unexpected error occurred during input plugin execution: {e}'
        )
        _terminate_subprocess(is_done, 1, queue)

    logger.info('Stopping input plugin')
    _terminate_subprocess(is_done, 0, queue)


@subprocess('event')
def start_event_subprocess(
    config: JinjaEventConfig,
    loader: BaseLoader | None,
    settings: Settings,
    input_queue: Queue,
    event_queue: Queue,
    is_done: EventClass
) -> None:
    logger.info('Initializing "jinja" event plugin')

    try:
        event_plugin = JinjaEventPlugin(config=config, loader=loader)
    except EventPluginConfigurationError as e:
        logger.error(f'Failed to initialize event plugin: {e}')
        _terminate_subprocess(is_done, 1, event_queue)
    except Exception as e:
        logger.error(
            f'Unexpected error occurred during initializing '
            f'event plugin: {e}'
        )
        _terminate_subprocess(is_done, 1, event_queue)

    logger.info('Event plugin is successfully initialized')

    timezone_as_string = datetime.now(
        tz=timezone(settings.timezone)
    ).strftime('%z')

    with Batcher(
        size=settings.output_batch_size,
        timeout=settings.output_batch_timeout,
        callback=event_queue.put
    ) as batcher:
        while True:
            timestamps_batch = input_queue.get()
            if timestamps_batch is None:
                break

            try:
                for timestamp in timestamps_batch:
                    for event in event_plugin.render(
                        **{
                            settings.timestamp_field_name: timestamp,
                            settings.timezone_field_name: timezone_as_string
                        }
                    ):
                        batcher.add(event)
            except EventPluginRuntimeError as e:
                logger.error(f'Failed to produce event: {e}')
                _terminate_subprocess(is_done, 1, event_queue)
            except Exception as e:
                logger.error(
                    f'Unexpected error occurred during producing event: {e}'
                )
                _terminate_subprocess(is_done, 1, event_queue)

    logger.info('Stopping event plugin')
    _terminate_subprocess(is_done, 0, event_queue)


@subprocess('output')
def start_output_subprocess(
    config: list[MutexFieldsModel],
    settings: Settings,
    queue: Queue,
    processed_events: SynchronizedBase,
    is_done: EventClass
) -> None:
    plugins_list_fmt = ", ".join(
        [f'"{item.get_name()}"' for item in config]
    )

    logger.info(f'Initializing [{plugins_list_fmt}] output plugins')

    output_plugins: list[BaseOutputPlugin] = []

    for item in config:
        plugin_name = item.get_name()
        output_conf = item.get_value()

        try:
            plugin_class = load_output_plugin_class(plugin_name=plugin_name)
            output_plugins.append(plugin_class(config=output_conf))
        except ValueError as e:
            logger.error(f'Failed to load output plugin: {e}')
            _terminate_subprocess(is_done, 1)
        except OutputPluginConfigurationError as e:
            logger.error(
                f'Failed to initialize "{plugin_name}" output plugin: {e}'
            )
            _terminate_subprocess(is_done, 1)
        except Exception as e:
            logger.error(
                'Unexpected error occurred during initializing '
                f'"{plugin_name}" output plugin: {e}'
            )
            _terminate_subprocess(is_done, 1)

    logger.info('Output plugins are successfully initialized')

    async def write_batch(
        plugin: BaseOutputPlugin,
        events_batch: NDArray[np.str_]
    ) -> None:
        batch_size = len(events_batch)
        try:
            if batch_size == 1:
                await plugin.write(events_batch[0])
            elif batch_size > 1:
                await plugin.write_many(events_batch)
        except OutputPluginRuntimeError as e:
            logger.error(f'Output plugin failed to write events: {e}')
            return

    async def run_loop() -> None:
        await asyncio.gather(
            *[plugin.open() for plugin in output_plugins]
        )

        while True:
            events_batch = queue.get()

            if events_batch is None:
                break

            await asyncio.gather(
                *[
                    write_batch(plugin, events_batch)
                    for plugin in output_plugins
                ]
            )

            processed_events.value += len(      # type: ignore[attr-defined]
                events_batch
            )

        await asyncio.gather(
            *[plugin.close() for plugin in output_plugins]
        )

    asyncio.run(run_loop())

    logger.info('Stopping output plugins')
    _terminate_subprocess(is_done, 0)



import time

from eventum_core.batcher import Batcher


def test_batcher_size_condition():
    bucket = []
    with Batcher(size=10, timeout=0.1, callback=bucket.append) as batcher:
        for i in range(100):
            batcher.add(i)

    assert len(bucket) == 10

    for batch in bucket:
        assert len(batch) == 10


def test_batcher_timeout_condition():
    bucket = []
    with Batcher(size=100, timeout=0.01, callback=bucket.append) as batcher:
        for i in range(10):
            batcher.add(i)
            time.sleep(0.05)

    assert len(bucket) == 10

    for batch in bucket:
        assert len(batch) == 1


def test_batcher_ordering():
    bucket = []
    with Batcher(size=100, timeout=1, callback=bucket.append) as batcher:
        for i in range(100):
            batcher.add(i)

    flattened_bucket = []
    for batch in bucket:
        flattened_bucket.extend(batch)

    flattened_bucket == list(range(100))



from enum import StrEnum
from typing import Any

from pydantic import BaseModel, Field, field_validator
from pytz import all_timezones_set


class TimeMode(StrEnum):
    SAMPLE = 'sample'
    LIVE = 'live'


class Settings(BaseModel, extra='forbid', frozen=True):
    # Time zone used in input plugins to generate timestamps.
    timezone: str = 'UTC'

    # The name of variable in template with timezone value (e.g. "+0300").
    timezone_field_name: str = 'tz'

    # The name of variable in template with original event timestamp.
    timestamp_field_name: str = Field('timestamp', min_length=1)

    # Batch size / timeout (in seconds) for input-to-event plugins
    # communication.
    events_batch_size: int = Field(1_000_000, ge=1)
    events_batch_timeout: float = Field(1.0, ge=0)

    # Batch size / timeout (in seconds) for output plugins.
    output_batch_size: int = Field(10_000, ge=1)
    output_batch_timeout: float = Field(1.0, ge=0)

    # Service name for keyring credentials storage
    keyring_service_name: str = Field('eventum', min_length=1)

    @field_validator('timezone')
    def validate_timezone(cls, v: Any):
        if v in all_timezones_set:
            return v

        raise ValueError(f'Unknown time zone "{v}"')


DEFAULT_SETTINGS = Settings()   # type: ignore[call-arg]





import time
from datetime import datetime
from multiprocessing import Event, Process, Queue, TimeoutError, Value

import numpy as np
from eventum_plugins.event.jinja import (JinjaEventConfig, TemplateConfig,
                                         TemplatePickingMode)
from eventum_plugins.input.sample import SampleInputConfig
from jinja2 import DictLoader

from eventum_core.app import TimeMode
from eventum_core.plugins_connector import InputConfigMapping
from eventum_core.settings import Settings
from eventum_core.subprocesses import (start_event_subprocess,
                                       start_input_subprocess,
                                       start_output_subprocess)


def test_input_subprocess():
    queue = Queue()
    is_done = Event()

    proc_input = Process(
        target=start_input_subprocess,
        args=(
            InputConfigMapping(sample=SampleInputConfig(count=10)),
            Settings(events_batch_timeout=0),
            TimeMode.SAMPLE,
            queue,
            is_done
        )
    )
    proc_input.start()

    try:
        is_done.wait(timeout=1)
    except TimeoutError:
        proc_input.terminate()
        raise AssertionError('Long subprocess execution')

    # IPC sleep
    time.sleep(0.5)

    assert not queue.empty()

    elements = []
    while True:
        batch = queue.get(timeout=0.1)

        if batch is None:
            break

        elements.extend(batch)

    assert len(elements) == 10

    proc_input.join()


def test_event_subprocess():
    input_queue = Queue()
    event_queue = Queue()
    is_done = Event()

    proc_event = Process(
        target=start_event_subprocess,
        args=(
            JinjaEventConfig(
                params={},
                samples={},
                mode=TemplatePickingMode.ALL,
                templates={
                    'test': TemplateConfig(template='test.jinja')
                }
            ),
            DictLoader({'test.jinja': '{{ timestamp }}'}),
            Settings(timestamp_field_name='timestamp', output_batch_timeout=0),
            input_queue,
            event_queue,
            is_done
        )
    )

    proc_event.start()
    input_queue.put(np.array([datetime.now().isoformat()] * 10))
    input_queue.put(None)

    try:
        is_done.wait(timeout=1)
    except TimeoutError:
        proc_event.terminate()
        raise AssertionError('Long subprocess execution')

    # IPC sleep
    time.sleep(0.5)

    assert not event_queue.empty()

    total_events = 0
    while True:
        batch = event_queue.get(timeout=0.1)

        if batch is None:
            break

        total_events += len(batch)

    assert total_events == 10

    proc_event.join()


def test_output_subprocess():
    event_queue = Queue()
    processed_event = Value('Q', 0)
    is_done = Event()

    proc_output = Process(
        target=start_output_subprocess,
        args=(
            [],
            Settings(),
            event_queue,
            processed_event,
            is_done
        )
    )

    proc_output.start()
    event_queue.put(np.array(['rendered event\n'] * 10))
    event_queue.put(None)

    try:
        is_done.wait(timeout=1)
    except TimeoutError:
        proc_output.terminate()
        raise AssertionError('Long subprocess execution')

    # IPC sleep
    time.sleep(0.5)

    assert processed_event.value == 10

    proc_output.join()



import logging
import signal
from multiprocessing import Event, Process, Queue, Value
from multiprocessing.sharedctypes import SynchronizedBase
from multiprocessing.synchronize import Event as EventClass
from time import sleep
from typing import NoReturn

import numpy as np
from eventum_plugins.event.jinja import JinjaEventConfig
from numpy.typing import NDArray
from pydantic import BaseModel
from setproctitle import getproctitle, setproctitle

from eventum_core.plugins_connector import (InputConfigMapping,
                                            OutputConfigMapping)
from eventum_core.settings import DEFAULT_SETTINGS, Settings, TimeMode
from eventum_core.subprocesses import (start_event_subprocess,
                                       start_input_subprocess,
                                       start_output_subprocess)

logger = logging.getLogger(__name__)


class ApplicationConfig(BaseModel, extra='forbid', frozen=True):
    input: InputConfigMapping           # type: ignore[valid-type]
    event: JinjaEventConfig
    output: list[OutputConfigMapping]   # type: ignore[valid-type]


class Application:
    """Main class of application."""

    _REFRESH_STATUS_INTERVAL = 0.1

    def __init__(
        self,
        config: ApplicationConfig,
        time_mode: TimeMode,
        settings: Settings = DEFAULT_SETTINGS,
    ) -> None:
        self._config = config
        self._time_mode = time_mode
        self._settings = settings

        # For all queues: The None element indicates that no more new
        # elements will be put in that queue
        self._input_queue: Queue[NDArray[np.datetime64]] = Queue()
        self._event_queue: Queue[NDArray[np.str_]] = Queue()

        # Regardless of whether the process ended with an error or not
        # this flag must be set at the end of its execution.
        # Used to control situations when process was killed from outside.
        self._is_input_done: EventClass = Event()
        self._is_event_done: EventClass = Event()
        self._is_output_done: EventClass = Event()

        self._processed_events: SynchronizedBase = Value('Q', 0)

        self._proc_input = Process(
            target=start_input_subprocess,
            args=(
                self._config.input,
                self._time_mode,
                self._settings,
                self._input_queue,
                self._is_input_done
            )
        )
        self._proc_event = Process(
            target=start_event_subprocess,
            args=(
                self._config.event,
                None,
                self._settings,
                self._input_queue,
                self._event_queue,
                self._is_event_done
            )
        )
        self._proc_output = Process(
            target=start_output_subprocess,
            args=(
                self._config.output,
                self._settings,
                self._event_queue,
                self._processed_events,
                self._is_output_done
            )
        )

    @property
    def processed_events(self) -> int:
        """Get currently processed events."""
        return self._processed_events.value     # type: ignore[attr-defined]

    def _terminate_application_on_crash(
        self,
        signal_number: int | None = None
    ) -> NoReturn:
        """Handle termination of application in emergency situation."""
        self._proc_input.terminate()
        self._proc_event.terminate()
        self._proc_output.terminate()

        if signal_number is not None:
            logger.info(
                f'Signal {signal.Signals(signal_number).name} is received'
            )

        logger.info('Application shut down')
        exit(1)

    def _register_signal_handlers(self) -> None:
        """Register handlers for received signals. Call this method
        only after starting all subprocesses to avoid inheritance of
        behavior in subprocesses."""
        signal.signal(
            signal.SIGINT,
            lambda signal, frame: self._terminate_application_on_crash(signal)
        )
        signal.signal(
            signal.SIGTERM,
            lambda signal, frame: self._terminate_application_on_crash(signal)
        )

    def start(self) -> None:
        logger.info('Application is started')

        self._proc_input.start()
        self._proc_event.start()
        self._proc_output.start()

        self._register_signal_handlers()

        setproctitle(f'{getproctitle()} [main]')

        while not self._is_output_done.is_set():
            if (
                not self._proc_input.is_alive()
                and not self._is_input_done.is_set()
            ):
                logger.critical(
                    'Input plugin subprocess terminated unexpectedly '
                    'or some error occurred'
                )
                logger.info('Application shut down')
                self._terminate_application_on_crash()

            if (
                not self._proc_event.is_alive()
                and not self._is_event_done.is_set()
            ):
                logger.critical(
                    'Event plugin subprocess terminated unexpectedly '
                    'or some error occurred'
                )
                logger.info('Application shut down')
                self._terminate_application_on_crash()

            if (
                not self._proc_output.is_alive()
                and not self._is_output_done.is_set()
            ):
                logger.critical(
                    'Output plugins subprocess terminated unexpectedly '
                    'or some error occurred'
                )
                logger.info('Application shut down')
                self._terminate_application_on_crash()

            sleep(Application._REFRESH_STATUS_INTERVAL)

        if self._proc_input.is_alive() or self._proc_event.is_alive():
            self._terminate_application_on_crash()
        else:
            self._proc_input.join()
            self._proc_event.join()
            self._proc_output.join()

        logger.info('Application shut down')
        exit(0)
